import logging

import pandas as pd

logger = logging.getLogger(__name__)

def trasform_to_numeric(df, verbose=logging.INFO):
    
    logger.setLevel(verbose)
    
    for col in df.columns:
        logger.info("\nProcessing: %s", col)
        
        is_numeric = pd.api.types.is_numeric_dtype(df[col])
        logger.info("is_numeric")

def remove_unneded_columns():
    
    # remove constants, etc
    pass


# read version from installed package
import logging
from importlib.metadata import version

logging.basicConfig(format='%(asctime)s | %(levelname)s | %(message)s', 
                    datefmt='%y-%m-%d %H:%M')

__version__ = version("instrumentum")




import logging

import numpy as np
import optuna
import sklearn
from joblib import Parallel, delayed
from sklearn.base import clone, is_classifier
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv, cross_val_score
from sklearn.model_selection._search import BaseSearchCV
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


class OptunaSearchCV(BaseSearchCV):

    _required_parameters = ["estimator", "search_space"]

    def __init__(
        self,
        estimator,
        search_space,
        n_iter=10,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=None,
        verbose=logging.INFO,
        random_state=None,
    ):
        self.search_space = search_space
        self.n_iter = n_iter
        self.random_state = random_state
        logger.setLevel(verbose)

        super().__init__(
            estimator=estimator,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=refit,
            cv=cv,
            verbose=0,
            pre_dispatch=None,
            error_score=np.nan,
            return_train_score=False,
        )

    def fit(self, X, y):

        # Validations
        if callable(self.scoring):
            scorer = self.scoring
        elif self.scoring is None or isinstance(self.scoring, str):
            scorer = check_scoring(self.estimator, self.scoring)
        else:
            # TODO: error
            raise "error"

        cv_orig = check_cv(self.cv, y, classifier=is_classifier(self.estimator))

        optuna.logging.set_verbosity(optuna.logging.ERROR)
        study = optuna.create_study(direction="maximize")
        study.optimize(
            lambda trial: self._opt_generic_objective(
                trial=trial,
                X=X,
                y=y,
                cv=cv_orig,
                scoring=scorer,
            ),
            n_trials=self.n_iter,
        )

        # Let's set some values so the upper classes can work
        if self.refit:
            self.best_estimator_ = clone(
                clone(self.estimator).set_params(**study.best_params)
            )
            self.best_estimator_.fit(X, y)

        self.scorer_ = scorer  # important so it knows how to calculate score
        self.study_ = study
        self.multimetric_ = None  # for compatibility

        return self

    def _opt_generic_objective(self, trial, X, y, cv, scoring):

        param = self.search_space(trial)
        candidate_estimator = clone(self.estimator)
        candidate_estimator.set_params(**param, random_state=self.random_state)

        score = cross_val_score(
            candidate_estimator, X=X, y=y, cv=cv, scoring=scoring
        ).mean()

        trial_n = len(trial.study.trials)
        best_score = (
            score
            if trial_n == 1 or score > trial.study.best_value
            else trial.study.best_value
        )

        logger.info("Trials: %s, Best Score: %s, Score %s", trial_n, best_score, score)
        return score

    # Note: parameters ending with "_" exists after fitting (sklearn logic follows that,
    # for example, check_is_fitted)

    @property
    def get_study_(self):
        check_is_fitted(self)
        return self.study_

    @property
    def best_score_(self):
        check_is_fitted(self)
        return self.study_.best_value

    @property
    def best_params_(self):
        check_is_fitted(self)
        return self.study_.best_params



import collections
import logging
import numbers

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import BaseEstimator, TransformerMixin, clone
from sklearn.exceptions import NotFittedError
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv
from sklearn.utils.validation import (
    FLOAT_DTYPES,
    _check_feature_names_in,
    check_is_fitted,
)

from instrumentum.utils._decorators import timeit
from instrumentum.utils.utils import check_jobs, get_combs

logger = logging.getLogger(__name__)


class Interactions(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        operations="prod",
        degree=2,
        check_features=False,
        estimator=None,
        verbose=logging.INFO,
    ):

        self.operations = operations
        self.degree = degree
        self.check_features = check_features
        self.estimator = estimator
        self.verbose = verbose

        logger.setLevel(verbose)

    def validate_input(self):

        if isinstance(self.operations, str):
            self.operations = [self.operations]

        if not isinstance(self.operations, list):
            raise ValueError("operations error input")

        if isinstance(self.degree, numbers.Integral):
            if self.degree < 0:
                raise ValueError(
                    f"degree must be a non-negative integer, got {self.degree}."
                )
            self._max_degree = self._min_degree = self.degree
        elif (
            isinstance(self.degree, collections.abc.Iterable) and len(self.degree) == 2
        ):
            self._min_degree, self._max_degree = self.degree
            if not (
                isinstance(self._min_degree, numbers.Integral)
                and isinstance(self._max_degree, numbers.Integral)
                and self._min_degree >= 0
                and self._min_degree <= self._max_degree
            ):
                raise ValueError(
                    "degree=(min_degree, max_degree) must "
                    "be non-negative integers that fulfil "
                    "min_degree <= max_degree, got "
                    f"{self.degree}."
                )
        else:
            raise ValueError(
                "degree must be a non-negative int or tuple "
                "(min_degree, max_degree), got "
                f"{self.degree}."
            )

    @timeit
    def fit(self, X, y=None):

        self.validate_input()

        X = self._validate_data(X)
        idxs = get_combs(
            self.n_features_in_, combs_from=self._min_degree, combs_to=self._max_degree
        )

        # I prefer a tuple due to its unmutability feature (to research)
        self.interactions_ = tuple((idx, op) for idx in idxs for op in self.operations)

        logger.info(
            "Total number of interactions calculated: %s", len(self.interactions_)
        )

    def get_interaction_name(self, idxs, op):

        feat_names = _check_feature_names_in(self).take(idxs)
        return str("_" + op + "_").join(feat_names)

    def transform(self, X, y=None):

        X = self._validate_data(X, order="F", dtype=FLOAT_DTYPES, reset=False)

        self.names_ = _check_feature_names_in(self)

        # To hold the interactions
        X_ = np.empty(shape=(len(X), len(self.interactions_)), dtype=X.dtype)
        # To hold the names of the interactions
        self.names_ = np.empty(shape=len(self.interactions_), dtype=object)

        for counter, (idxs, op) in enumerate(self.interactions_):
            res = self.get_op_result(X, idxs, op)
            X_[:, counter] = res
            self.names_[counter] = self.get_interaction_name(idxs, op)

            logger.debug("New feature created: %s", self.names_[counter])

        self.names_ = np.hstack(
            (
                _check_feature_names_in(self),
                self.names_,
            )
        )

        return np.hstack((X, X_))

    def get_feature_names_out(self):
        return self.names_

    def get_op_result(self, arr, idxs, op):
        if op == "sum":
            return arr.take(idxs, axis=1).sum(axis=1)

        if op == "prod":
            return arr.take(idxs, axis=1).prod(axis=1)



import inspect
import logging
import time

logger = logging.getLogger(__name__)

def timeit(func):

    def wrap_func(*args, **kwargs):
        t1 = time.time()
        result = func(*args, **kwargs)
        t2 = time.time()
  
        verbose = None
        
        if len(args) > 0 and hasattr(args[0], 'verbose'):
            verbose = args[0].verbose
        
        elif 'verbose' in kwargs:
            verbose = kwargs.get('verbose', wrap_func.__name__.upper())

        verbose and logger.setLevel(verbose)
        
        logger.info('Function %s executed in %s seconds', func.__name__, (t2-t1))
        return result
    
    return wrap_func




import numpy as np
import pandas as pd

from instrumentum.utils._decorators import timeit


def _min_corr_groups(df: pd.DataFrame, threshold: float = 0.8) -> pd.Series:
   
    corr = df.corr().abs()
    corr.loc[:,:] =  np.tril(corr, k=-1)
    corr = corr.unstack()
    corr = corr[corr >= threshold]
    corr.sort_values(inplace=True, ascending=True)
    return [set(x) for x in corr.index]

def _can_add_item(item, group, corr_pairs):

    if item in group:
        return False

    all_combs = [{item,x} for x in group]

    if all(any(y.issubset(x) for x in corr_pairs) for y in all_combs):
        return True

def _remove_redundancy(l):
    # this line removes the sets that are equal
    l = [set(item) for item in set(frozenset(item) for item in l)]

    # this removes all elements included in other (if I dont execute the previous line)
    # then sets that are equal are both removed
    return [x for x in l if not any(x<=y for y in l if x is not y)]


def _get_correlated_items(item, corr_pairs):
    
    return set().union(*[x for x in corr_pairs if item in x])



def get_corr_groups_accurate(df, threshold):

    corr_pairs = _min_corr_groups(df, threshold)

    # The initial groups will be the minimum expression
    corr_groups = corr_pairs.copy()

    # These are the columns we need, let's avoid to process the ones
    # that do not have any pair over the threshold
    columns = set().union(*corr_pairs)

    keep_going = True
    while(keep_going):

        new_groups = []
        for group in corr_groups:
            for col in columns:
                # If the item can be added to an existing group, then I will not add
                # it directly to that specific groupo, but rather will create another entry with them
                # the reason is that another variable can come in later and do not fit in becuase of the 
                # recently added item, thus we will leave the original group for every variable to try.
                # later we will remove redundancy.
                if _can_add_item(col, group, corr_pairs):
                    new_groups.append(group | set([col]))

        if(len(new_groups)!=0):
            corr_groups += new_groups           
            corr_groups = _remove_redundancy(corr_groups)
        else:
            keep_going = False

    return corr_groups


def get_corr_simple(df: pd.DataFrame, threshold: float):
    # This is not deterministic

    # Get list of all pairs correlated (based on threshold)
    corr_pairs = _min_corr_groups(df, threshold)
    
    corr_groups = []

    # These are the columns we need, let's avoid to process the ones
    # that do not have any pair correlated
    columns = set().union(*corr_pairs)

    for col in columns:
        added_items = set().union(*corr_groups)

        if(col not in added_items):
            correlated_items = _get_correlated_items(col, corr_pairs)

            correlated_items.difference_update(added_items)

            if(len(correlated_items)>1): # 0 or itself no append
                corr_groups.append(correlated_items)

        
    return corr_groups



import logging

import optuna
import optuna.integration.lightgbm as lgb
from catboost import CatBoostClassifier
from lightgbm import LGBMClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBRegressor


def _xgbclassifier_default(trial: optuna.trial.Trial):
    param = {

        "objective": "binary:logistic",
        "eval_metric": "auc",
        "booster": trial.suggest_categorical("booster", ["gbtree", "gblinear", "dart"]),
        "lambda": trial.suggest_loguniform("lambda", 1e-3, 10),
        "alpha": trial.suggest_loguniform("alpha", 1e-3, 10),
        "learning_rate": trial.suggest_float("learning_rate", 1e-5, 1e-1, log=True),
        "n_estimators": trial.suggest_int("n_estimators", 20, 600, step=20),
    }

    if param["booster"] == "gbtree" or param["booster"] == "dart":
        param["max_depth"] = trial.suggest_int("max_depth", 1, 9)
        param["eta"] = trial.suggest_loguniform("eta", 1e-8, 1.0)
        param["gamma"] = trial.suggest_loguniform("gamma", 1e-8, 1.0)
        param["grow_policy"] = trial.suggest_categorical(
            "grow_policy", ["depthwise", "lossguide"]
        )
    if param["booster"] == "dart":
        param["sample_type"] = trial.suggest_categorical(
            "sample_type", ["uniform", "weighted"]
        )
        param["normalize_type"] = trial.suggest_categorical(
            "normalize_type", ["tree", "forest"]
        )
        param["rate_drop"] = trial.suggest_loguniform("rate_drop", 1e-8, 1.0)
        param["skip_drop"] = trial.suggest_loguniform("skip_drop", 1e-8, 1.0)

    return param


def _lgbmclassifier_default(trial: optuna.trial.Trial):
    params = {
        #"verbosity": -1,
        "boosting_type": trial.suggest_categorical(
            "boosting_type", ["gbdt", "dart", "goss"]
        ),
        "verbose": -1,
        "objective": "binary",
        "metric": ["binary", "binary_error", "auc"],
        "num_leaves": trial.suggest_int("num_leaves", 5, 500, step=5),
        "max_depth": trial.suggest_int("max_depth", 1, 15),
        "learning_rate": trial.suggest_float("learning_rate", 1e-5, 1.0, log=True),
        "colsample_bytree": trial.suggest_uniform("colsample_bytree", 0.0, 1.0),
        'reg_alpha': trial.suggest_loguniform('reg_alpha', 1e-3, 30),
        'reg_lambda': trial.suggest_loguniform('reg_lambda', 1e-3, 30),
        "min_split_gain": trial.suggest_float("min_split_gain", 0, 15),
    }
    # if params["boosting_type"] == "dart":
    #     params["drop_rate"] = trial.suggest_loguniform("drop_rate", 1e-8, 1.0)
    #     params["skip_drop"] = trial.suggest_loguniform("skip_drop", 1e-8, 1.0)
    # if params["boosting_type"] == "goss":
    #     params["top_rate"] = trial.suggest_uniform("top_rate", 0.0, 1.0)
    #     params["other_rate"] = trial.suggest_uniform(
    #         "other_rate", 0.0, 1.0 - params["top_rate"]
    #     )

    return params


def _catboostclassifier_default(trial: optuna.trial.Trial):
    params = {
        "allow_writing_files": False,
        "logging_level": 'Silent',
        #"silent": True,
        #"verbose": 0,
        "iterations": trial.suggest_int("iterations", 50, 300),
        "depth": trial.suggest_int("depth", 2, 10),
        "learning_rate": trial.suggest_loguniform("learning_rate", 0.01, 0.3),
        "random_strength": trial.suggest_int("random_strength", 0, 100),
        "bagging_temperature": trial.suggest_loguniform(
            "bagging_temperature", 0.01, 100.00
        ),
        "od_type": trial.suggest_categorical("od_type", ["IncToDec", "Iter"]),
        "od_wait": trial.suggest_int("od_wait", 10, 50),
    }

    return params


def _random_forest_classifier_default(trial: optuna.trial.Trial):
    params = {
        "n_estimators": trial.suggest_int("n_estimators", 10, 120),
        "max_depth": trial.suggest_int("max_depth", 1, 12)
    }

    return params

def _decision_tree_classifier_default(trial: optuna.trial.Trial):
    params = {
        "max_depth": trial.suggest_int("max_depth", 1, 12),
        "criterion": trial.suggest_categorical("criterion", ["gini", "entropy"])
    }

    return params


def _decision_tree_regressor_default(trial: optuna.trial.Trial):
    params = {
        "max_depth": trial.suggest_int("max_depth", 1, 12),
      #  "criterion": trial.suggest_categorical("criterion", ["gini", "entropy"])
    }

    return params

optuna_param_disp = {
    XGBRegressor.__name__: _xgbclassifier_default,
    LGBMClassifier.__name__: _lgbmclassifier_default,
    RandomForestClassifier.__name__: _random_forest_classifier_default,
    CatBoostClassifier.__name__: _catboostclassifier_default,
    DecisionTreeClassifier.__name__: _decision_tree_classifier_default,
    DecisionTreeRegressor.__name__: _decision_tree_regressor_default,
}



import logging
import multiprocessing
from itertools import chain, combinations

logger = logging.getLogger(__name__)


def get_combs(set_size, combs_to, combs_from=1):

    if combs_from > combs_to or combs_from < 0:
        raise ValueError("combs_to must be possitive and less or equal to combs_from")
    # We accepts combs_to > set_size, it will just use set_size as the limit

    l_comb = [combinations(range(set_size), x) for x in range(combs_from, combs_to + 1)]

    return tuple(chain.from_iterable(l_comb))


def check_jobs(n_jobs, verbose=logging.INFO):

    logger.setLevel(verbose)

    verified_jobs = n_jobs
    max_jobs = multiprocessing.cpu_count()

    if n_jobs > max_jobs:
        logger.warning(
            "Max cores in this coputer %s, lowering to that from input %s",
            max_jobs,
            n_jobs,
        )
        verified_jobs = max_jobs
    elif n_jobs < -1:
        logger.error(
            "Jobs cannot be negative (except for -1 for all cores). Your total cores \
            are %s , modifying your entry to 1",
            max_jobs,
        )
        verified_jobs = 1

    else:
        if n_jobs == -1:
            verified_jobs = max_jobs

        logger.info(
            "Number of cores to be used: %s, total available: %s\n",
            verified_jobs,
            max_jobs,
        )

    return verified_jobs



import logging
from itertools import chain, combinations

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone
from sklearn.exceptions import NotFittedError
from sklearn.feature_selection import SelectorMixin
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv
from sklearn.utils.validation import _check_feature_names_in, check_is_fitted

from instrumentum.utils._decorators import timeit
from instrumentum.utils.utils import check_jobs, get_combs

logger = logging.getLogger(__name__)

# TODO:
# receive a function for scoring. And add the scoring as a method
# agregar una historia de casos para que no vuelva a verificar casos ya verificados
# Document the full shit
class DynamicStepwise(SelectorMixin, MetaEstimatorMixin, BaseEstimator):
    def __init__(
        self,
        estimator,
        *,
        n_combs=1,
        rounding=4,
        add_always=False,
        direction="forward",
        n_jobs=-1,
        verbose=logging.INFO,
        # Max cols doesnt mean exactly this number, but
        # no more than this number (unless add_always is true)
        max_cols=None,
    ):

        self.estimator = estimator
        self.n_combs = n_combs
        self.rounding = rounding
        self.add_always = add_always
        self.direction = direction
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.max_cols = max_cols

        logger.setLevel(verbose)

    @timeit
    def fit(self, X, y):

        if self.direction not in ("forward", "backward"):
            raise ValueError(
                "direction must be either 'forward' or 'backward'. "
                f"Got {self.direction}."
            )

        # features_name_in are set if X is a DataFrame
        X, y = self._validate_data(
            X,
            y,
        )

        n_features = X.shape[1]
        self.n_jobs = check_jobs(self.n_jobs, self.verbose)
        self.seq_cols_added_ = []
        self._candidates_history = {}
        # the current mask corresponds to the set of features:
        # - that we have already *selected* if we do forward selection
        # - that we have already *excluded* if we do backward selection
        current_mask = np.zeros(n_features, dtype=bool)

        if self.direction == "backward":
            est = clone(self.estimator)
            est.fit(X[:, ~current_mask], y)
            tracker_score = round(est.best_score_, self.rounding)
            logger.info("With all columns, score is %s\n", tracker_score)
        else:
            # For forward, we know we have to at least add 1 column
            tracker_score = 0

        keep_going = True

        while keep_going:

            n_cols_remaining = sum(~current_mask)
            n_combs_ = self._get_n_combs(current_mask)
            combs = get_combs(set_size=n_cols_remaining, combs_to=n_combs_)

            logger.info("Remaining columns to test: %s", n_cols_remaining)
            logger.info("Combinations to test: %s", len(combs))

            comb_results = Parallel(n_jobs=self.n_jobs)(
                delayed(self._get_best_columns)(
                    X,
                    y,
                    estimator=clone(self.estimator),
                    rounding=self.rounding,
                    comb=comb,
                    mask=current_mask,
                )
                for comb in combs
            )

            keep_going, tracker_score, current_mask = self._update_process(
                comb_results, tracker_score, current_mask
            )

        self.mask_ = current_mask if self.direction == "forward" else ~current_mask

    def _get_n_combs(self, current_mask):
        if not self.max_cols:
            return self.n_combs
        # If max columns is set, and the next iterations only allow n
        # more columns, but combs > n, this will avoid testing more
        # than what max cols dictate
        n_combs_ = min(self.n_combs, self.max_cols - sum(current_mask))

        # Finishing conditions should finish the process before this could happen
        assert n_combs_ > 0

        return n_combs_

    def _update_process(self, comb_results, tracker_score, current_mask):

        assert len(comb_results) > 0
        # easier to query
        fwd = self.direction == "forward"

        # Add to the history of candidates
        self._candidates_history.update(
            {
                frozenset(np.append(comb, np.flatnonzero(current_mask))): score
                for score, comb in comb_results
            }
        )

        # Get the best ccandidate
        best_comb_score, idx_best_comb_cols = max(
            comb_results, key=lambda item: item[0]
        )

        logger.info(
            "Best score from combinations: %s, global score %s",
            best_comb_score,
            tracker_score,
        )

        logger.info(
            "Best score comes from %s columns: %s",
            "adding" if fwd else "removing",
            self._get_all_features_in()[idx_best_comb_cols],
        )

        # When forward, we don't want to add if they dont improve the score (>)
        # for backward, we prefer keep removing if they don't reduce the score (>=)
        if (
            best_comb_score > tracker_score if fwd else best_comb_score >= tracker_score
        ) or self.add_always:

            current_mask[idx_best_comb_cols] = True
            tracker_score = best_comb_score
            self.seq_cols_added_.append(
                (best_comb_score, self._get_all_features_in()[idx_best_comb_cols])
            )

            logger.info(
                "Best columns were %s. All columns %s so far %s\n",
                "added" if fwd else "removed",
                "added" if fwd else "removed",
                self._get_all_features_in()[current_mask],
            )
            keep_going = True

        else:
            logger.info(
                "Columns were not %s as they do not improve the score. Finishing\n",
                "added" if fwd else "removed",
            )
            keep_going = False

        # Finish conditions because of limits
        if all(current_mask):
            logger.info(
                "All columns were %s. Finishing.\n", "added" if fwd else "removed"
            )
            keep_going = False

        elif self.max_cols and current_mask.sum() >= self.max_cols:
            logger.info("Max columns %s reached. Finishing.\n", self.max_cols)
            keep_going = False

        return keep_going, tracker_score, current_mask

    def _get_best_columns(self, X, y, estimator, rounding, comb, mask):
        logger.setLevel(self.verbose)  # needed for delayed

        try:
            check_is_fitted(estimator)
        except NotFittedError:
            pass
        else:
            raise ValueError("Estimator was fitted already")

        idx_not_processed = np.flatnonzero(~mask)
        idx_comb_to_eval = np.take(idx_not_processed, comb)

        if self.direction == "forward":
            mask_candidate = mask.copy()
        else:
            mask_candidate = ~mask.copy()

        # To be comptaible with both forward and backward
        mask_candidate[idx_comb_to_eval] = not all(mask_candidate[idx_comb_to_eval])

        # Did we calculated this already?
        saved_score = self._candidates_history.get(
            frozenset(np.flatnonzero(mask_candidate)), None
        )

        if saved_score:
            score = saved_score
        else:
            if sum(mask_candidate) == 0:
                # For backward, it might remove all columns for some candidates, and won't fit.
                score = 0
            else:
                estimator.fit(X[:, mask_candidate], y)
                score = round(estimator.best_score_, rounding)

        logger.debug(
            "Score %s checking %s full combination: %s %s",
            score,
            self._get_all_features_in()[idx_comb_to_eval],
            self._get_all_features_in()[mask_candidate],
            "* already calculated" if saved_score else "",
        )

        return score, idx_comb_to_eval

    # this is an utility function to get the names of the
    # columns passed
    # it works with numpy and dataframes (leverage some
    # sklearn functionality).
    def _get_all_features_in(self):
        return _check_feature_names_in(self)

    def _get_support_mask(self):
        return self.mask_

    @property
    def seq_columns_selected_(self):
        # todo if fitted...
        # How to act on this when returned (asuming ret), options:
        # (1) pd.DataFrame(ret, columns=['a', 'b']) # Pretty printer
        # (2) np.concatenate([x[1] for x in ret]) # Get all columns
        return self.seq_cols_added_



import logging

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def get_col_nans(df, cutoff=0):
    # nan_rows = df[df.isnull().T.any()]
    perc = df.isnull().mean()

    return perc.loc[perc > cutoff]


def get_col_non_numeric(df):
    return df.select_dtypes(exclude=[np.number]).dtypes



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['instrumentum',
 'instrumentum.analysis',
 'instrumentum.feature_generation',
 'instrumentum.feature_preprocess',
 'instrumentum.feature_selection',
 'instrumentum.image_processing',
 'instrumentum.model_tuning',
 'instrumentum.time_series',
 'instrumentum.utils']

package_data = \
{'': ['*']}

install_requires = \
['catboost>=1.0.0,<2.0.0',
 'joblib>=1.1.0,<2.0.0',
 'lightgbm>=3.3.0,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'optuna>=2.10.0,<3.0.0',
 'pandas>=1.3.3,<2.0.0',
 'sklearn>=0.0,<0.1',
 'xgboost>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'instrumentum',
    'version': '0.8.0',
    'description': 'General utilities for data science projects',
    'long_description': '# instrumentum\n\nGeneral utilities for data science projects\n\n## Installation\n\n```bash\n$ pip install instrumentum\n```\n\n## Usage\n\n- TODO\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`instrumentum` was created by Federico Montanana. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`instrumentum` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n',
    'author': 'Federico Montanana',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import logging

import numpy as np
from scipy.cluster.hierarchy import dendrogram, fcluster, linkage
from scipy.spatial.distance import squareform
from scipy.stats import kendalltau, spearmanr
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone
from sklearn.exceptions import NotFittedError
from sklearn.feature_selection import SelectorMixin
from sklearn.metrics import check_scoring
from sklearn.model_selection import check_cv
from sklearn.utils.validation import _check_feature_names_in, check_is_fitted

from instrumentum.utils._decorators import timeit
from instrumentum.utils.utils import check_jobs

logger = logging.getLogger(__name__)


def idx_from_mask(mask):
    return np.flatnonzero(mask)  # ~ to negate


def mask_from_idx(idx, size):
    mask = np.zeros(size, dtype=bool)
    mask[idx] = True
    return mask


class ClusterSelection(SelectorMixin, MetaEstimatorMixin, BaseEstimator):
    def __init__(
        self,
        method: str = "pearson",
        t: float = 0.8,
        meta_estimator=None,
        verbose=logging.INFO,
        make_matrix=None,
        criterion="distance",
    ):

        self.t = t
        self.method = method
        self.meta_estimator = meta_estimator
        self.verbose = verbose
        self.make_matrix = make_matrix
        self.criterion = criterion

        logger.setLevel(verbose)

    # TODO: agregar las opciones de distancia etc.
    def _get_clusters(self, X, y, as_dict=True):

        if self.make_matrix is not None:
            dis = self.make_matrix(X, y)
        else:
            X_corr = np.corrcoef(X, rowvar=False)  # spearmanr(X)[0]
            dis = 1 - np.fabs(X_corr)
        # shuldnt be necessary but sometimes it is not zero
        # and we get an error in squareforms
        np.fill_diagonal(dis, 0, wrap=False)
        dis = np.maximum(dis, dis.transpose())
        #
        Z = linkage(squareform(dis), "complete")

        clusters = fcluster(Z, self.t, criterion=self.criterion)

        if not as_dict:
            return clusters

        return {k: np.where(clusters == k)[0] for k in set(clusters)}

    def fit(self, X, y=None):

        X = self._validate_data(X)

        constant_mask = self._check_constants(X)
        if any(constant_mask):
            raise ValueError(
                "You must remove the constant columns before using this functionality"
            )

        self.clusters_in_ = self._get_clusters(X, y)
        self.clusters_out_ = {}

        current_mask = np.zeros(X.shape[1], dtype=bool)

        for id_cluster, id_cols in self.clusters_in_.items():

            if len(id_cols) == 1:
                logger.info(
                    "Cluster %s - Unique column: %s. Selected Automatically\n",
                    id_cluster,
                    self._get_all_features_in()[id_cols],
                )
                best_idx = id_cols
            else:
                logger.info(
                    "Cluster %s - Evaluating columns: %s",
                    id_cluster,
                    self._get_all_features_in()[id_cols],
                )

                best_idx = self._get_best_from_cluster(X, y, id_cols)

                logger.info(
                    "Columns selected: %s\n", self._get_all_features_in()[best_idx]
                )

            current_mask |= mask_from_idx(best_idx, X.shape[1])
            self.clusters_out_[id_cluster] = best_idx

        self.mask_ = current_mask

    def _get_best_from_cluster(self, X, y, id_cols):

        est = clone(self.meta_estimator)
        est.fit(X[:, id_cols], y)
        best_idx = est.get_support(indices=True)

        return id_cols[best_idx]

    def _check_constants(self, X):
        peak_to_peaks = np.ptp(X, axis=0)
        return np.isclose(peak_to_peaks, 0.0)

    def _get_all_features_in(self):
        return _check_feature_names_in(self)

    def _get_support_mask(self):
        return self.mask_

    @property
    def clusters_initial(self):
        # todo if fitted...
        return self.clusters_in_

    @property
    def clusters_final(self):
        # todo if fitted...
        return self.clusters_out_


# if __name__ == "__main__":
#     from collections import Counter

#     from sklearn.datasets import make_classification
#     from sklearn.tree import DecisionTreeClassifier

#     from instrumentum.feature_selection.correlation import ClusterSelection
#     from instrumentum.feature_selection.stepwise import DynamicStepwise
#     from instrumentum.model_tuning._optuna_dispatchers import optuna_param_disp
#     from instrumentum.model_tuning.wrapper_optuna import OptunaSearchCV
#     from sklearn.model_selection import train_test_split
#     import logging

#     import numpy as np

#     from sklearn.metrics import roc_auc_score
#     from lightgbm import LGBMClassifier

#     X, y = make_classification(
#         n_samples=2000,
#         n_redundant=5,
#         n_features=10,
#         n_informative=5,
#         random_state=111,
#     )

#     search_space = optuna_param_disp[DecisionTreeClassifier.__name__]
#     os = OptunaSearchCV(
#         estimator=DecisionTreeClassifier(),
#         scoring="roc_auc",
#         search_space=search_space,
#         n_iter=5,
#         verbose=logging.WARNING,
#         random_state=123,
#     )

#     stepw = DynamicStepwise(
#         estimator=os,
#         rounding=4,
#         n_combs=2,
#         verbose=logging.WARNING,
#         direction="forward",
#         max_cols=2,
#     )

#     threshold = 0.8

#     def score_me(X, y):
#         search_space = optuna_param_disp[LGBMClassifier.__name__]
#         os = OptunaSearchCV(
#             estimator=LGBMClassifier(),
#             scoring="roc_auc",
#             search_space=search_space,
#             n_iter=20,
#             verbose=logging.WARNING,
#             random_state=123,
#         )

#         X_train, X_test, y_train, _ = train_test_split(
#             X, y, test_size=0.25, random_state=42, stratify=y
#         )

#         return os.fit(X_train.reshape(-1, 1), y_train).predict(X_test.reshape(-1, 1))

#     def create_corr(X, y):

#         n = X.shape[1]
#         correl = np.ones((n, n), dtype=float)

#         scores = {}

#         for i in range(n - 1):
#             for j in range(i + 1, n):

#                 if i not in scores:
#                     scores[i] = score_me(X[:, i], y)

#                 if j not in scores:
#                     scores[j] = score_me(X[:, j], y)

#                 c_i = np.isclose(np.ptp(scores[i], axis=0), 0)
#                 c_j = np.isclose(np.ptp(scores[j], axis=0), 0)

#                 r = 0.5
#                 if not c_j:
#                     r = roc_auc_score(scores[j], scores[i])
#                 elif not c_i:
#                     r = roc_auc_score(scores[i], scores[j])

#                 correl[i, j] = correl[j, i] = r

#         return abs(correl - 0.5) * 2

#     # custom_corr_matrix = create_corr(X, y)
#     # custom_corr_matrix = abs(custom_corr_matrix - 0.5) * 2

#     clsl = ClusterSelection(
#         threshold=0.05, meta_estimator=stepw, make_matrix=create_corr
#     )
#     clsl.fit(X, y)

#     print("\nThese are the best columns after removal: ")
#     print(clsl.get_feature_names_out())

#     print("\nNumber of elements grouped by clusters before reduction: ")
#     print(Counter([len(v) for k, v in clsl.clusters_in_.items()]))

#     print("\nNumber of elements grouped by clusters after reduction: ")
#     print(Counter([len(v) for k, v in clsl.clusters_out_.items()]))



import logging

import optuna
import optuna.integration.lightgbm as lgb
import pandas as pd
from catboost import CatBoostClassifier
from lightgbm import LGBMClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import (RepeatedStratifiedKFold, StratifiedKFold,
                                     cross_val_score)
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier, XGBRegressor, plot_importance

from instrumentum.model_tuning._optuna_dispatchers import optuna_param_disp

logger = logging.getLogger(__name__)


def _opt_generic_objective(X, y, trial, estimator, cv, metric):

    param = optuna_param_disp[estimator.__name__](trial)
    estimator = estimator(**param)

    score = cross_val_score(estimator, X=X, y=y, cv=cv, scoring=metric).mean()

    trial_n = len(trial.study.trials)
    best_score = (
        score
        if trial_n == 1 or score > trial.study.best_value
        else trial.study.best_value
    )

    logger.info("Trials: %s, Best Score: %s, Score %s", trial_n, best_score, score)
    return score


def wrapper_opt(
    X,
    y,
    estimator=None,
    metric="roc_auc",
    n_trials=5,
    verbose=logging.INFO,
    return_fit=True,
    direction="maximize",
    cv_splits=5,
    cv_repeats=1,
):
    # Our Logger
    logger.setLevel(verbose)
    # Let's turn off the verbosity of optuna
    optuna.logging.set_verbosity(optuna.logging.ERROR)

    cv = RepeatedStratifiedKFold(n_splits=cv_splits, n_repeats=cv_repeats)
    estimator = estimator or DecisionTreeClassifier

    logger.info("Estimator received: %s, trials: %s\n", estimator.__name__, n_trials)

    study = optuna.create_study(direction=direction)
    study.optimize(
        lambda trial: _opt_generic_objective(
            trial=trial,
            X=X,
            y=y,
            estimator=estimator,
            cv=cv,
            metric=metric,
        ),
        n_trials=n_trials,
    )

    estimator = estimator(**study.best_params)
    return_fit and estimator.fit(X, y)

    return study.best_trial.value, estimator


def wrapper_opt_lgbm(
    X, y, metric="auc", time_budget=120, verbose=logging.INFO, return_fit=False
):

    # Our Logger
    logger.setLevel(verbose)

    # Let's turn off the verbosity of optuna and lighgbm
    optuna.logging.set_verbosity(optuna.logging.ERROR)

    no_logger = logging.getLogger("sd")
    no_logger.addHandler(logging.NullHandler())
    lgb.register_logger(no_logger)

    def log_trials(std, frz_trial):
        logger.info(
            "\nTrials: %s, Iteration Score: %s", len(std.trials), std.best_value
        )

    params = {
        "objective": "binary",
        "metric": metric,
        "boosting_type": "gbdt",
        "seed": 42,
    }

    dtrain = lgb.Dataset(X, label=y)
    rkf = RepeatedStratifiedKFold(
        n_splits=10,
        n_repeats=2,
        random_state=42,
    )
    study_tuner = optuna.create_study(direction="maximize")

    tuner = lgb.LightGBMTunerCV(
        params,
        dtrain,
        study=study_tuner,
        time_budget=time_budget,
        seed=42,
        optuna_callbacks=[log_trials],
        show_progress_bar=False,
        folds=rkf,
    )

    tuner.run()

    lgbm = LGBMClassifier(**tuner.best_params)
    return_fit and lgbm.fit(X, y)

    return tuner.best_score, lgbm




    
def ts_print(df, col_target, col_key, value_true=1):
    """ Print a Pandas dataframe with the required format

    Parameters
    ----------
    df : Dataframe
        A dataframe in the required timeframe form
    col_target : [type]
        [description]
    col_key : [type]
        [description]
    value_true : int, optional
        [description], by default 1
    """
    
    
    # Let's get all the customers that had at least one open card (drop duplicates; same cust can appear more than once)
    keys_pos_unique = df.loc[df[col_target] == value_true][col_key].drop_duplicates(
        keep="first"
    )

    # Let's get all customers who neer had a card opened (Since cust can be in many datasets, drop duplicates)
    keys_neg_unique = df.loc[~df[col_key].isin(keys_pos_unique)][
        col_key
    ].drop_duplicates(keep="first")

    print("Total unique keys that with a possitive value: ", len(keys_pos_unique))
    print(
        "Total records of ckeys that with a possitive value: ",
        df.loc[df[col_key].isin(keys_pos_unique)].shape[0],
    )

    print("Total unique keys with no possitive value: ", len(keys_neg_unique))
    print(
        "Total records of customers with no possitive values: ",
        df.loc[df[col_key].isin(keys_neg_unique)].shape[0],
    )
